@@ -82,6 +82,9 @@ def get_routes(telescope: Telescope) -> dict:
     dc_routes_get = {
         '/datalogging/currentState': lambda args: telescope.datalogging_currentState(*args),
         '/devices/statusValue': lambda args: telescope.devices_statusValue(args),
+        '/devices/getAllDeviceStatusValues': lambda args: telescope.devices_getAllDeviceStatusValues(args),
+        '/devices/statusPaths': lambda args: telescope.devices_statusPaths(args),
+
         '/datalogging/lastSession': lambda args: telescope.datalogging_lastSession(*args),
         '/datalogging/exportSession': lambda args: telescope.datalogging_exportSession(args),
         '/datalogging/sessions': lambda args: telescope.datalogging_sessions(*args),
@@ -134,7 +137,7 @@ class scu_sim(scu.scu):
 
     (This class inherits from a real SCU interface object and overwrites the communication channels)
     """
-    def __init__(self, ip='localhost', port='8080', use_realtime=False, debug=True, speedup_factor=1, t_start = astropy.time.Time.now(), UPDATE_INTERVAL = .1):
+    def __init__(self, address='', ip='', port='8080', use_realtime=False, debug=True, speedup_factor=1, t_start = astropy.time.Time.now(), UPDATE_INTERVAL = .1, lims_az=(-270.0, 270, 3.0), lims_el=(15, 90, 1.35), dish_type='mke', post_put_delay=0.0, **kwargs):
         """create an antenna scu/acu simulator object, which can be used to simulate the motion of
         a real Antenna in simulation time (ahead of time) in order to debug test scripts etc.
 
@@ -145,6 +148,8 @@ class scu_sim(scu.scu):
             ip (str, optional): NOT USED FOR SIMULATOR: ip address of the antenna to connect to. Defaults to 'localhost'.
             port (str, optional): NOT USED FOR SIMULATOR: port of the antenna to connect to. Defaults to '8080'.
             debug (bool, optional): Set to True, to receive additional information in stdout, when using commands. Defaults to True.
+            lims_az (tuple, optional): limits on AZ axis, angle_min, angle_max, speed_max. Defaults to (-270.0, 270, 3.0).
+            lims_el (tuple, optional): limits on EL axis, angle_min, angle_max, speed_max. Defaults to (15, 90, 1.35).
         """
 
         self.dc = {}
@@ -163,8 +168,69 @@ class scu_sim(scu.scu):
         self.routes = get_routes(self.telescope)
 
 
-        scu.scu.__init__(self, ip=ip, port=port, debug=debug)
+        scu.scu.__init__(self, address=address, ip=ip, port=port, debug=debug, lims_az=lims_az, lims_el=lims_el, dish_type=dish_type, post_put_delay=post_put_delay, use_socket=False)
+    
 
+    @property
+    def data_streamer(self):
+        return None
+
+    @data_streamer.setter
+    def data_streamer(self, value):
+        pass
+
+    @property
+    def use_socket(self):
+        return None
+    
+    @use_socket.setter
+    def use_socket(self, value):
+        pass
+
+    @property
+    def data(self):
+        return self.telescope.data
+
+    def getc(self, key, as_dict=False):
+        """get one or many channel values. key(s) must be string or list of strings
+
+        Args:
+            key (string or iterable[str]): the channel names to get
+
+        Returns:
+            dict, list or immuteable: type depending on the input.
+        """
+        try:
+            if self.use_socket:
+                if isinstance(key, str):
+                    return {key: self[key]} if as_dict else self[key]
+                else:
+                    return self.get_many(channels=key, as_dict=as_dict)
+        except Exception as err:
+            log(f'error while getting data from socket stream... falling back to HTTP api. {err=}', color=colors.WARNING)
+
+        if key is None:
+            return dict(self.get_channel_list(with_values=True, with_timestamps=False))
+        elif not isinstance(key, str) and hasattr(self, '__len__'):
+            allchans = self.data
+            if as_dict:
+                return {k:allchans[k] for k in key}
+            else:
+                return [allchans[k] for k in key]
+        else:
+            return self._get_device_status_value(key)
+
+    def __getitem__(self, key):
+        if not isinstance(key, str) and hasattr(key, '__len__'):
+            return [self.data[k] for k in key]
+        else:
+            return self.data[key]
+    
+    def __contains__(self, key):
+        if not isinstance(key, str) and hasattr(key, '__len__'):
+            return True if [k for k in key if k in self.data] else False
+        else:
+            return key in self.data
 
     @property
     def t_internal(self):
@@ -182,7 +248,7 @@ class scu_sim(scu.scu):
         with params=payload (OVERWRITTEN FOR SIMULATION!)'''
         URL = 'http://' + self.ip + ':' + self.port + device
         if device != '/devices/statusValue':
-            self.event_log[self.telescope.t_internal.datetime] = 'GET | ' + device
+            self.call_log[self.telescope.t_internal.datetime] = 'GET | ' + device
 
         if device not in self.routes['GET']:
             r = MockResponseObject(URL, 404, {})
@@ -199,7 +265,7 @@ class scu_sim(scu.scu):
         '''This is a generic PUT command into http: scu port + folder 
         with json=payload (OVERWRITTEN FOR SIMULATION!)'''
         URL = 'http://' + self.ip + ':' + self.port + device
-        self.event_log[self.telescope.t_internal.datetime] = 'PUT | ' + device
+        self.call_log[self.telescope.t_internal.datetime] = 'PUT | ' + device
         if device not in self.routes['PUT']:
             r = MockResponseObject(URL, 404, {})
         else:
@@ -207,6 +273,10 @@ class scu_sim(scu.scu):
             res = fun(payload, params, data)
             r = MockResponseObject(URL, res['status'], res['body'])
         self._feedback(r)
+        
+        if self.post_put_delay > 0:
+            self.wait_duration(self.post_put_delay, no_stout=True)
+        
         return(r)
 
     def scu_delete(self, device, payload = {}, params = {}):
@@ -255,6 +325,32 @@ class scu_sim(scu.scu):
     #     log('   -> done')
 
 
+    def wait_settle(self, axis='all', timeout=600, query_delay=.25, tolerance=0.01, wait_by_pos=False, initial_delay=1.0):
+        """
+        alias for waitForStatusValue but mapping 'AZ', 'EL', 'FI' to 'acu.azimuth.p_act'
+        'acu.elevation.p_act' and 'acu.feed_indexer.p_act'
+
+        Periodically queries a device status 'path' until a specific value is reached.
+
+        Args:
+            path:       path of the SCU device status
+            Value:      value to be reached
+            timeout:    Raise TimeoutError after this duration
+            query_delay: Period in seconds to wait between two queries.
+        """
+        
+        if initial_delay > 0:
+            # assures setpoint has actually been send to acu!
+            self.wait_duration(initial_delay, no_stout=True)  
+
+        if axis == 'all':          
+            self.wait_settle('az', initial_delay=0.0, wait_by_pos=True)
+            self.wait_settle('el', initial_delay=0.0, wait_by_pos=True)
+            self.wait_settle('fi', initial_delay=0.0)
+            return
+        else:
+            super().wait_settle(axis=axis, timeout=timeout, query_delay=query_delay, tolerance=tolerance, wait_by_pos=wait_by_pos, initial_delay=0.0)
+
     def get_history_df(self, interval_ms = None):
         """get this simulators internal motion history as DataFrame
 
@@ -288,7 +384,7 @@ class scu_sim(scu.scu):
         self.get_command_authority()
         # self.reset_dmc()
         self.wait_duration(3)
-        self.unstow()
+        self.unstow(nowait=True)
 
         self.wait_duration(5)
         # self.activate_dmc()
@@ -312,7 +408,7 @@ class scu_sim(scu.scu):
         """Stow, deactivate, and release command authority for antenna in order to finish before handing back the antenna
         """
         log('=== INITIATING SHUTDOWN ROUTINE ===')
-        self.stow()
+        self.stow(nowait=True)
         self.wait_duration(5)
         # self.deactivate_axes()
         self.wait_duration(5)
@@ -394,7 +490,7 @@ def plot_motion_pyplot(df, xkey='index', figsize=(12, 10), df_tt=None, event_log
         ax.plot(x, df[actp], '-' + c)
 
         if df_tt is not None and s in df_tt:
-            ax.plot(x_tt, df_tt[s], ':k')
+            ax.plot(x_tt, df_tt[s], ':k', label=f'tracking table: {s}')
 
         if xlims is None:
             xlims = ax.get_xlim()
